@@ -8,17 +8,22 @@ There are 4 levels of constructs.
   and uses that single process only. (If the code futher creates other processes,
   well, no one can forbid that, but that is not the intended usage.)
 
+  There are two worker classes: `ProcessWorker` and `ThreadWorker`.
+
 - A `Servlet` manages a `Worker`, specifing how many processes are to be created,
   each executing the worker code independently. It may control exactly
   which CPU(s) each worker process uses. The servlet code runs in the "main process",
   whereas workers run in other processes. Each input item is processed by one
-  and only one worker. If `W1` is a subclass of `Worker`, then the simplest servlet
+  and only one worker.
+
+  There are two servelet classess: `ProcessServlet` and `ThreadServlet`.
+
+  If `W1` is a subclass of `ProcessWorker`, then the simplest servlet
   is like this:
 
-    s = Servlet(W1)
+    s = ProcessServlet(W1)
 
-  This creates as many processes as there are CPU cores, minus core 0, and runs the worker
-  code in each process.
+  This creates one worker process that is not pinned to a particular CPU.
 
 - Servlets can be composed in `Sequential`s or `Ensemble`s. In a `Sequential`,
   the first servlet's output becomes the second servlet's input, and so on.
@@ -27,22 +32,25 @@ There are 4 levels of constructs.
 
   Great power comes from the fact that both `Sequential` and `Ensemble`
   also follow the `Servlet` API, hence both can be constituents of other compositions.
-  In principle, you can freely compose and nest them (in practice, you can't go crazy;
-  remember they create processes). For example, suppose `W1`, `W2`,..., are `Worker`
-  subclasses, then you may design such a workflow,
+  In principle, you can freely compose and nest them.
+  For example, suppose `W1`, `W2`,..., are `Worker` subclasses,
+  then you may design such a workflow,
 
     s = Sequential(
-            Servlet(W1),
+            ProcessServlet(W1),
             Ensemble(
-                Servlet(W2),
-                Sequential(Servlet(W3), Servlet(W4)),
+                ThreadServlet(W2),
+                Sequential(ProcessServlet(W3), ThreadServlet(W4)),
                 ),
             Ensemble(
-                Sequetial(Servlet(W5), Servlet(W6)),
-                Sequetial(Servlet(W7), Servlet(W8), Servlet(W9)),
+                Sequetial(ProcessServlet(W5), ProcessServlet(W6)),
+                Sequetial(ProcessServlet(W7), ThreadServlet(W8), ProcessServlet(W9)),
                 ),
         )
 
+  In sum, `ProcessServlet`, `ThreadServlet`, `Sequential`, `Ensemble` are collectively
+  referred to and used as `Servlet`.
+
 - On the top level is `Server`. Pass a `Servlet`, or `Sequential` or `Ensemble`
   into a `Server`, which handles scheduling as well as interfacing with the outside
   world:
@@ -68,7 +76,7 @@ CPU allocations among workers to achieve best performance.
 Reference: [Service Batching from Scratch, Again](https://zpz.github.io/blog/batched-service-redesign/).
 This article describes roughly version 0.7.2.
 '''
-
+from __future__ import annotations
 import asyncio
 import concurrent.futures
 import logging
@@ -77,14 +85,15 @@ import multiprocessing.queues
 import queue
 import threading
 from abc import ABCMeta, abstractmethod
+from datetime import datetime
 from queue import Empty
 from time import perf_counter, sleep
-from typing import Sequence, Union, Callable, Type, Protocol, Any
+from typing import Sequence, Union, Callable, Type, Any
 
 import psutil
 
 from .remote_exception import RemoteException, exit_err_msg
-from .util import is_exception, Thread, Process, ProcessLogger
+from .util import is_exception, Thread, Process, ProcessLogger, TimeoutError
 from ._queues import SingleLane
 
 
@@ -96,10 +105,6 @@ logger = logging.getLogger(__name__)
 NOMOREDATA = b'c7160a52-f8ed-40e4-8a38-ec6b84c2cd87'
 
 
-class TimeoutError(Exception):
-    pass
-
-
 class FastQueue(multiprocessing.queues.SimpleQueue):
     # Check out os.read, os.write, os.close with file-descriptor args.
     def __init__(self, *, ctx):
@@ -109,19 +114,32 @@ class FastQueue(multiprocessing.queues.SimpleQueue):
         self._rlock = ctx.RLock()
 
 
+class SimpleQueue(queue.SimpleQueue):
+    def __init__(self):
+        super().__init__()
+        self._rlock = threading.RLock()
+
+
 class Worker(metaclass=ABCMeta):
     # Typically a subclass needs to enhance
     # `__init__` and implement `call`.
 
     @classmethod
-    def run(cls, *, q_in: FastQueue, q_out: FastQueue, **init_kwargs):
+    def run(cls, *,
+            q_in: Union[FastQueue, SimpleQueue],
+            q_out: Union[FastQueue, SimpleQueue],
+            **init_kwargs):
+        # For a `ProcessWorker`, `q_in` and `q_out` should be `FastQueue`.
+        # For a `ThreadWorker`, `q_in` and `q_out` are either `FastQueue` or `SimpleQueue`.
         obj = cls(**init_kwargs)
         q_out.put(obj.name)
+        # This sends a signal to caller indicating completion of init.
         obj.start(q_in=q_in, q_out=q_out)
 
     def __init__(self, *,
                  batch_size: int = None,
-                 batch_wait_time: float = None):
+                 batch_wait_time: float = None,
+                 batch_size_log_cadence: int = 10_000):
         '''
         `batch_size`: max batch size; see `call`.
 
@@ -169,6 +187,7 @@ class Worker(metaclass=ABCMeta):
                 assert 0 < batch_wait_time < 1
 
         self.batch_size = batch_size
+        self.batch_size_log_cadence = batch_size_log_cadence
         self.batch_wait_time = batch_wait_time
         self.name = multiprocessing.current_process().name
 
@@ -213,26 +232,19 @@ class Worker(metaclass=ABCMeta):
     def start(self, *, q_in, q_out):
         try:
             if self.batch_size > 1:
-                self._batch_buffer = SingleLane(self.batch_size + 10)
-                self._batch_get_called = threading.Event()
-                t = Thread(target=self._build_input_batches, args=(q_in, q_out))
-                t.start()
                 self._start_batch(q_in=q_in, q_out=q_out)
             else:
                 self._start_single(q_in=q_in, q_out=q_out)
         except KeyboardInterrupt:
             print(self.name, 'stopped by KeyboardInterrupt')
-        finally:
-            if self.batch_size > 1:
-                t.join()
+            # The process will exit. Don't print the usual
+            # exception stuff as that's not need when user
+            # pressed Ctrl-C.
 
     def _start_single(self, *, q_in, q_out):
         batch_size = self.batch_size
         while True:
-            try:
-                z = q_in.get()
-            except Empty:
-                continue
+            z = q_in.get()
             if z == NOMOREDATA:
                 q_out.put(z)
                 q_in.put(z)  # broadcast to one fellow worker
@@ -261,6 +273,11 @@ class Worker(metaclass=ABCMeta):
             logger.info('%d batches with sizes %d--%d, mean %.1f',
                         n_batches, batch_size_min, batch_size_max, batch_size_mean)
 
+        self._batch_buffer = SingleLane(self.batch_size + 10)
+        self._batch_get_called = threading.Event()
+        collector_thread = Thread(target=self._build_input_batches, args=(q_in, q_out))
+        collector_thread.start()
+
         n_batches = 0
         try:
             while True:
@@ -293,12 +310,14 @@ class Worker(metaclass=ABCMeta):
                 batch_size_max = max(batch_size_max, n)
                 batch_size_min = min(batch_size_min, n)
                 batch_size_mean = (batch_size_mean * (n_batches - 1) + n) / n_batches
-                if n_batches % 1000 == 0:
-                    print_batching_info()
-                    n_batches = 0
+                if self.batch_size_log_cadence:
+                    if n_batches % self.batch_size_log_cadence == 0:
+                        print_batching_info()
+                        n_batches = 0
         finally:
             if n_batches:
                 print_batching_info()
+            collector_thread.join()
 
     def _build_input_batches(self, q_in, q_out):
         threading.current_thread().name = f"{self.name}._build_input_batches"
@@ -416,25 +435,66 @@ class ProcessWorker(Worker):
             log_passer.stop()
 
 
-class ServletProtocol(Protocol):
-    def start(self, q_in, q_out, *, log_passer, ctx):
-        pass
-
-    def stop(self):
-        pass
+class ThreadWorker(Worker):
+    '''
+    Use this class if the operation is I/O bound (e.g. calling an external service
+    to get some info), and computation is very light compared to the I/O part.
+    Another use-case of this class is to perform some very simple and quick
+    pre-processing or post-processing.
+    '''
+    pass
 
 
-class Servlet:
+def make_threadworker(func: Callable[[Any], Any]):
+    # Define a simple ThreadWorker subclass for quick, "on-the-fly" use.
+    # This can be useful when we want to introduce simple servlets
+    # for pre-processing and post-processing.
+    class MyThreadWorker(ThreadWorker):
+        def call(self, x):
+            return func(x)
+
+    MyThreadWorker.__name__ = f"ThreadWorker-{func.__name__}"
+    return MyThreadWorker
+
+
+PassThrough = make_threadworker(lambda x: x)
+# Example use of this class:
+#
+#    def combine(x):
+#       '''
+#       Combine the ensemble elements depending on the results
+#       as well as the original input.
+#       '''
+#       x, *y = x
+#       assert len(y) == 3
+#       if x < 100:
+#           return sum(y) / len(y)
+#       else:
+#           return max(y)
+#
+#   s = Ensemble(
+#           ThreadServlet(PassThrough),
+#           ProcessServlet(W1),
+#           ProcessServlet(W2)
+#           ProcessServlet(W3),
+#       )
+#   ss = Sequential(s, ThreadServlet(make_threadworker(combine)))
+
+
+class ProcessServlet:
     def __init__(self,
                  worker_cls: Type[ProcessWorker],
                  *,
                  cpus: list = None,
-                 workers: int = None,
                  name: str = None,
                  **kwargs):
+        '''
+        `cpus`: specifies how many processes to create and how they are pinned
+                to specific CPUs. The default is a single unpinned process.
+        '''
         self._worker_cls = worker_cls
         self._name = name or worker_cls.__name__
-        self._cpus = self._resolve_cpus(cpus=cpus, workers=workers)
+        self._cpus = cpus
         self._init_kwargs = kwargs
         self._workers = []
         self._started = False
@@ -444,9 +504,24 @@ class Servlet:
         `ctx` is a multiprocessing context passed in ultimately from `Server`.
         '''
         assert not self._started
-        for cpu in self._cpus:
-            # Pinned to the specified cpu core.
-            sname = f"{self._name}-{','.join(map(str, cpu))}"
+
+        # CPU allocation can look as flexible as this:
+        #   [[0, 1, 2], [0], [2, 3], [4, 5, 6], 4, None]
+        cpus = self._cpus or [None]
+        for i, c in enumerate(cpus):
+            if c is not None:
+                if isinstance(c, int):
+                    cpus[i] = [c]
+                else:
+                    assert all(isinstance(v, int) for v in c)
+
+        for cpu in cpus:
+            # Create as many processes as the length of `cpus`.
+            # Each process is pinned to the specified cpu core.
+            if cpu is None:
+                sname = self._name
+            else:
+                sname = f"{self._name}-{','.join(map(str, cpu))}"
             logger.info('adding worker <%s> at CPU %s ...', sname, cpu)
             w = Process(
                 ctx=ctx,
@@ -464,6 +539,7 @@ class Servlet:
             w.start()
             name = q_out.get()
             logger.debug(f"   ... worker <{name}> is ready")
+
         logger.info(f"servlet {self._name} is ready")
         self._started = True
 
@@ -474,44 +550,49 @@ class Servlet:
         self._workers = []
         self._started = False
 
-    def _resolve_cpus(self, *, cpus: list = None, workers: int = None):
-        n_cpus = psutil.cpu_count(logical=True)
 
-        # Either `workers` or `cpus`, but not both,
-        # can be specified.
-        if workers:
-            # Number of workers is specified.
-            # Create this many processes, each pinned
-            # to one core. One core will host mulitple
-            # processes if `workers` exceeds the number
-            # of cores.
-            assert not cpus
-            assert 0 < workers <= n_cpus * 4
-            cpus = list(reversed(range(n_cpus))) * 4
-            cpus = sorted(cpus[:workers])
-            # List[int]
-        elif cpus:
-            assert isinstance(cpus, list)
-            # Create as many processes as the length of `cpus`.
-            # Each element of `cpus` specifies cpu pinning for
-            # one process. `cpus` could contain repeat numbers,
-            # meaning multiple processes can be pinned to the same
-            # cpu.
-            # This provides the ultimate flexibility, e.g.
-            #    [[0, 1, 2], [0], [2, 3], [4, 5, 6], 4, None]
-        else:
-            # Create as many processes as there are cores minus core '0',
-            # one process pinned to one core.
-            cpus = list(range(1, n_cpus))
-            # List[int]
+class ThreadServlet:
+    def __init__(self,
+                 worker_cls: Type[ThreadWorker],
+                 *,
+                 num_threads: int = None,
+                 name: str = None,
+                 **kwargs):
+        self._worker_cls = worker_cls
+        self._name = name or worker_cls.__name__
+        self._num_threads = num_threads or 1
+        self._init_kwargs = kwargs
+        self._workers = []
+        self._started = False
 
-        for i, cpu in enumerate(cpus):
-            if isinstance(cpu, int):
-                cpu = [cpu]
-                cpus[i] = cpu
-            assert all(0 <= c < n_cpus for c in cpu)
+    def start(self, q_in, q_out):
+        assert not self._started
+        for ithread in range(self._num_threads):
+            sname = f"{self._name}-{ithread}"
+            logger.info('adding worker <%s> in thread ...', sname)
+            w = Thread(
+                target=self._worker_cls.run,
+                name=sname,
+                kwargs={
+                    'q_in': q_in,
+                    'q_out': q_out,
+                    **self._init_kwargs,
+                },
+            )
+            self._workers.append(w)
+            w.start()
+            name = q_out.get()
+            logger.debug(f"   ... worker <{name}> is ready")
 
-        return cpus
+        logger.info(f"servlet {self._name} is ready")
+        self._started = True
+
+    def stop(self):
+        assert self._started
+        for w in self._workers:
+            w.join()
+        self._workers = []
+        self._started = False
 
 
 class Sequential:
@@ -520,7 +601,7 @@ class Sequential:
     the previous op's output becoming the subsequent
     op's input.
     '''
-    def __init__(self, *servlets: ServletProtocol):
+    def __init__(self, *servlets: Servlet):
         assert len(servlets) > 0
         self._servlets = servlets
         self._qs = []
@@ -532,11 +613,17 @@ class Sequential:
         q1 = q_in
         for i, s in enumerate(self._servlets):
             if i + 1 < nn:
-                q2 = FastQueue(ctx=ctx)
+                if isinstance(s, ThreadServlet) and isinstance(self._servlets[i + 1], ThreadServlet):
+                    q2 = SimpleQueue()
+                else:
+                    q2 = FastQueue(ctx=ctx)
                 self._qs.append(q2)
             else:
                 q2 = q_out
-            s.start(q1, q2, log_passer=log_passer, ctx=ctx)
+            if isinstance(s, ThreadServlet):
+                s.start(q1, q2)
+            else:
+                s.start(q1, q2, log_passer=log_passer, ctx=ctx)
             q1 = q2
         self._started = True
 
@@ -547,6 +634,13 @@ class Sequential:
         self._qs = []
         self._started = False
 
+    @property
+    def _workers(self):
+        w = []
+        for s in self._servlets:
+            w.extend(s._workers)
+        return w
+
 
 class Ensemble:
     '''
@@ -554,17 +648,9 @@ class Ensemble:
     in parallel, the list of results are gathered and combined
     to form a final result.
     '''
-    def __init__(self, *servlets: ServletProtocol,
-                 post_func: Callable[[Any, list], Any] = None):
-        '''
-        `post_func`: if provided, this function takes the original input
-            value, `x`, and the ensemble output components, `y`
-            (a list containing output of each servlet, in order),
-            and returns a final result; if absent, `y` is the output.
-        '''
+    def __init__(self, *servlets: Servlet):
         assert len(servlets) > 1
         self._servlets = servlets
-        self._func = post_func
         self._started = False
 
     def _reset(self):
@@ -581,9 +667,14 @@ class Ensemble:
         self._qin = q_in
         self._qout = q_out
         for s in self._servlets:
-            q1 = FastQueue(ctx=ctx)
-            q2 = FastQueue(ctx=ctx)
-            s.start(q1, q2, log_passer=log_passer, ctx=ctx)
+            if isinstance(s, ThreadServlet):
+                q1 = SimpleQueue()
+                q2 = SimpleQueue()
+                s.start(q1, q2)
+            else:
+                q1 = FastQueue(ctx=ctx)
+                q2 = FastQueue(ctx=ctx)
+                s.start(q1, q2, log_passer=log_passer, ctx=ctx)
             self._qins.append(q1)
             self._qouts.append(q2)
         t = Thread(target=self._dequeue)
@@ -600,7 +691,6 @@ class Ensemble:
         qout = self._qout
         qins = self._qins
         catalog = self._uid_to_results
-        hasfunc = self._func is not None
         nn = len(qins)
         while True:
             v = qin.get()
@@ -614,15 +704,12 @@ class Ensemble:
                 qout.put((uid, x))
                 continue
             z = {'y': [None] * nn, 'n': 0}
-            if hasfunc:
-                z['x'] = x
             catalog[uid] = z
             for q in qins:
                 q.put((uid, x))
 
     def _dequeue(self):
         threading.current_thread().name = f"{self.__class__.__name__}._dequeue"
-        func = self._func
         qout = self._qout
         qouts = self._qouts
         catalog = self._uid_to_results
@@ -644,8 +731,6 @@ class Ensemble:
                         # All results for this request have been collected.
                         catalog.pop(uid)
                         y = z['y']
-                        if func is not None:
-                            y = func(z['x'], y)
                         qout.put((uid, y))
 
     def stop(self):
@@ -657,9 +742,26 @@ class Ensemble:
         self._reset()
         self._started = False
 
+    @property
+    def _workers(self):
+        w = []
+        for s in self._servlets:
+            w.extend(s._workers)
+        return w
+
+
+Servlet = Union[ProcessServlet, ThreadServlet, Sequential, Ensemble]
+
 
 class Server:
-    def __init__(self, servlet: ServletProtocol, *, main_cpu: int = 0, backlog: int = 1024, ctx=None):
+    def __init__(self,
+                 servlet: Servlet,
+                 *,
+                 main_cpu: int = 0,
+                 backlog: int = 1024,
+                 ctx=None,
+                 sys_info_log_cadence: int = 10_000,
+                 ):
         '''
         `main_cpu`: specifies the cpu for the "main process",
         i.e. the process in which this server objects resides.
@@ -668,6 +770,9 @@ class Server:
             all pipes/servlets/stages combined.
 
         `ctx`: a multiprocessing context. Leave this at default unless you know what you're doing.
+
+        `sys_info_log_cadence`: log worker process system info (cpu/memory utilization)
+            every this many requests; if `None`, do not log.
         '''
         self._servlet = servlet
 
@@ -688,6 +793,7 @@ class Server:
                 cpus = main_cpu
             psutil.Process().cpu_affinity(cpus=cpus)
 
+        self._sys_info_log_cadence = sys_info_log_cadence
         self._started = False
 
     def __enter__(self):
@@ -713,17 +819,18 @@ class Server:
 
         self._q_in = None
         self._q_out = None
-        if isinstance(self._servlet, Ensemble):
+        if isinstance(self._servlet, (Ensemble, ThreadServlet)):
             self._q_in = queue.Queue()
             self._q_out = queue.Queue()
         else:
             self._q_in = FastQueue(ctx=self.get_mpcontext())
             self._q_out = FastQueue(ctx=self.get_mpcontext())
-        self._servlet.start(
-            self._q_in,
-            self._q_out,
-            log_passer=self._worker_logger,
-            ctx=self.get_mpcontext())
+        if isinstance(self._servlet, ThreadServlet):
+            self._servlet.start(self._q_in, self._q_out)
+        else:
+            self._servlet.start(
+                self._q_in, self._q_out,
+                log_passer=self._worker_logger, ctx=self.get_mpcontext())
 
         t = Thread(target=self._gather_output)
         t.start()
@@ -944,6 +1051,35 @@ class Server:
         q_out = self._q_out
         futures = self._uid_to_futures
 
+        log_cadence = self._sys_info_log_cadence
+        if log_cadence:
+            ts0 = datetime.utcnow()
+            logcounter = 0
+
+            def _log_sys_info():
+                pp = [(w.name, psutil.Process(w.pid))
+                      for w in self._servlet._workers
+                      if not isinstance(w, Thread)]
+                msg = [f"  time from           {ts0}",
+                       f"  time to             {datetime.utcnow()}",
+                       f"  items served        {logcounter}",
+                       ]
+                attrs = ['memory_full_info',
+                         'cpu_affinity', 'cpu_percent', 'cpu_times',
+                         'num_threads', 'num_fds', 'io_counters',
+                         'status',
+                         ]
+                for pname, pobj in pp:
+                    msg.append(f'    process {pname}')
+                    info = pobj.as_dict(attrs)
+                    for k in attrs:
+                        v = info[k]
+                        if k == 'memory_full_info':
+                            msg.append(f"      {'memory_uss':<15} {v.uss / 1_000_000:.2f} MB")
+                        else:
+                            msg.append(f"      {k:<15} {v}")
+                logger.info('worker process stats:\n%s', '\n'.join(msg))
+
         while True:
             z = q_out.get()
             if z == NOMOREDATA:
@@ -965,3 +1101,14 @@ class Server:
                     # Unexpected situation; to be investigated.
                     logger.exception(e)
                     raise
+
+            if log_cadence:
+                logcounter += 1
+                if logcounter >= log_cadence and (q_out.empty() or logcounter >= log_cadence * 10):
+                    _log_sys_info()
+                    logcounter = 0
+                    ts0 = datetime.utcnow()
+
+        if log_cadence:
+            if logcounter >= log_cadence / 2:
+                _log_sys_info()
