@@ -3,10 +3,11 @@ import concurrent.futures
 import logging
 import os
 import queue
+import stat
 import threading
 import time
 from pickle import dumps as pickle_dumps, loads as pickle_loads
-from time import monotonic
+from time import perf_counter
 from typing import Iterable, Union, Sequence, Callable, Awaitable, Any
 
 from overrides import EnforceOverrides
@@ -106,15 +107,10 @@ async def run_unix_server(conn_handler: Callable, path: str):
         and it (including parent directories) does not need to exist.
         In fact, if the file exists, it will be removed first.
     '''
-    try:
-        os.unlink(path)
-    except FileNotFoundError:
+    if os.path.exists(path):
+        assert stat.S_ISSOCK(os.stat(path).st_mode), f"file '{path}' exists but is not a socket file"
+    else:
         os.makedirs(os.path.dirname(path), exist_ok=True)
-    except PermissionError:
-        raise
-    except OSError:
-        if os.path.exists(path):
-            raise
     server = await asyncio.start_unix_server(conn_handler, path)
     async with server:
         addrs = ', '.join(str(sock.getsockname()) for sock in server.sockets)
@@ -123,14 +119,14 @@ async def run_unix_server(conn_handler: Callable, path: str):
 
 
 async def open_tcp_connection(host, port, *, timeout=None):
-    t0 = monotonic()
+    t0 = perf_counter()
     while True:
         try:
             reader, writer = await asyncio.open_connection(host, port)
             return reader, writer
         except ConnectionRefusedError:
             if timeout is not None:
-                if monotonic() - t0 > timeout:
+                if perf_counter() - t0 > timeout:
                     raise
             await asyncio.sleep(0.1)
 
@@ -139,14 +135,14 @@ async def open_unix_connection(path: str, *, timeout=None):
     '''
     `path`: the same string that has been used by `run_unix_server`.
     '''
-    t0 = monotonic()
+    t0 = perf_counter()
     while True:
         try:
             reader, writer = await asyncio.open_unix_connection(path)
             return reader, writer
         except (ConnectionRefusedError, FileNotFoundError):
             if timeout is not None:
-                if monotonic() - t0 > timeout:
+                if perf_counter() - t0 > timeout:
                     raise
             await asyncio.sleep(0.1)
 
@@ -230,7 +226,7 @@ class SocketServer(EnforceOverrides):
         if path:
             assert not host
             assert not port
-            self._path = path
+            self._path = os.path.abspath(path)
             self._host = None
             self._port = None
         else:
@@ -466,13 +462,13 @@ class SocketClient(EnforceOverrides):
             logger.error(msg)
 
         self._prepare_shutdown.set()
-        t0 = monotonic()
+        t0 = perf_counter()
         while not self._pending_requests.empty():
-            if monotonic() - t0 > self._shutdown_timeout:
+            if perf_counter() - t0 > self._shutdown_timeout:
                 break
             time.sleep(0.1)
         while self._active_requests:
-            if monotonic() - t0 > self._shutdown_timeout:
+            if perf_counter() - t0 > self._shutdown_timeout:
                 break
             time.sleep(0.1)
         self._to_shutdown.set()
@@ -607,7 +603,7 @@ class SocketClient(EnforceOverrides):
         If `return_x` is `False`, return a stream of `y`.
         If `return_exceptions` is `True`, `y` could be an Exception object.
         '''
-        # Refer to `mpserver.MPServer.stream` for in-code documentation.
+        # Refer to `mpserver.Server.stream` for in-code documentation.
         tasks = SingleLane(self._backlog)
         nomore = object()
 
@@ -627,7 +623,7 @@ class SocketClient(EnforceOverrides):
                     else:
                         logger.error("exception '%r' happened for input '%s'", e, x)
                         raise
-                t0 = monotonic()
+                t0 = perf_counter()
                 if not put_in_queue(tt, (x, fut, t0), to_shutdown):
                     return
             put_in_queue(tt, nomore, to_shutdown)
@@ -652,7 +648,7 @@ class SocketClient(EnforceOverrides):
                 break
             x, fut, t0 = z
             try:
-                y = fut.result(timeout=response_timeout - (monotonic() - t0))
+                y = fut.result(timeout=response_timeout - (perf_counter() - t0))
             except Exception as e:
                 if return_exceptions:
                     if return_x:
